/* Vector utils: a series of functions for dealing with vectors
 *
 * Author: Dylan Muir (dr.muir@student.qut.edu.au)
 *         QUT MLRC LPG Semester 2 1999
 * Date: 14th Septamber, 1999
 * Modified: 13th July, 2000
 * Version: 0.12
 */

#ifndef __vector_utils_h
#define	__vector_utils_h

/* -- Vector includes */

#include "stddefs.h"

/* -- Vector structures */

typedef struct {
	int		dimensions;
	double	*components;
} vector;


/* -- Vector functions */

/* -- Function VectorAllocate
 * Pre: 'dimensions' is the number of dimensions to allocate for this vector
 * Post: (The vector was allocated && returned) ||
 *       (there was an error allocating the vector && NULL was returned)
 * Note: a new vector has undefined components
 */
vector *VectorAllocate(int dimensions);

/* -- Function VectorConstruct
 * Pre: 'dimensions' is the number of dimensions to allocate for this vector
 *      'components' is a double array of components to place in this vector
 * Post: (A new vector was allocated based on 'components', the contents of
 *        'components' was copied into the vector && the vector was returned) ||
 *       (There was an error allocating the vector && NULL was returned)
 */
vector *VectorConstruct(int dimensions, const double *components);

/* -- Function VectorDeallocate
 * Pre: 'vect' is a valid allocated vector
 * Post: 'vect' was deallocated && NULL was returned
 */
vector *VectorDeallocate(vector *vect);

/* -- Function VectorEuclideanDist
 * Pre: 'vect1' and 'vect2' are valid allocated vectors
 *      the dimensionality of 'vect1' and 'vect2' is the same
 */
double VectorEuclideanDist(vector *vect1, vector *vect2);

/* -- Function VectorLength
 * Pre: 'vect' is a valid allocated vector
 * Post: returned the euclidean distance from 0 to 'vect'
 */
double VectorLength(vector *vect);

/* -- Function VectorZero
 * Pre: 'dimensions' > 0
 * Post: (returned a reference to a zero vector) ||
 *       (there was an error in allocation && returned NULL)
 * Note: VectorZero allocates a new vector, which you MUST deallocate!
 */
vector *VectorZero(int dimensions);

/* -- Function VectorUnit
 * Pre: 'dimensions' > 0
 * Post: (returned a reference to a unit vector in 'dimensions' dimensions) ||
 *       (there was an error in allocation && returned NULL)
 * Note: VectorZero allocates a new vector, which you MUST deallocate!
 */
vector *VectorUnit(int dimensions);

/* -- Function VectorDotProduct
 * Pre: 'vect1' and 'vect2' are valid allocated vectors
 *      the dimensionality of 'vect1' and 'vect2' is equal
 * Post: returned the dot product of 'vect1' and 'vect2'
 * Note: The name for this function is a little ambiguous;
 *       all the functions in the vector suite are prepended
 *       with 'Vector', but this function performs the
 *       _scalar_ dot product of two vectors.  There's no
 *       thing as a vector dot product, is there?
 */
double VectorDotProduct(vector *vect1, vector *vect2);

/* -- Function VectorSum
 * Pre: 'vectd' and 'vects' are valid allocated vectors
 *      the dimensionality of 'vectd' and 'vects' is the same
 * Post: 'vectd' = 'vectd' + 'vects'
 */
void VectorSum(vector *vectd, vector *vects);

/* -- Function VectorSub
 * Pre: 'vectd' and 'vects' are valid allocated vectors
 *      the dimensionality of 'vectd' and 'vects' is the same
 * Post: 'vectd' = 'vectd' - 'vects'
 */
void VectorSub(vector *vectd, vector *vects);

/* -- Function VectorNeg
 * Pre: 'vect' is a valid allocated vector
 * Post: 'vect' = -'vect'
 */
void VectorNeg(vector *vect);

/* -- Function VectorSquare
 * Pre: 'vect' is a valid allocated vector
 * Post: arg(i) 'vect'i = 'vect'i * 'vect'i
 *       i.e. each element of the vector is squared
 * Note: Performs x.^2
 */
void VectorSquare(vector *vect);

/* -- Function VectorMultiplyScalar
 * Pre: 'vect' is a valid allocated vector
 *      'scalar' is a valid number
 * Post: 'vect' = 'vect' * 'scalar'
 */
void VectorMultiplyScalar(vector *vect, double scalar);

/* -- Function VectorDivideScalar
 * Pre: 'vect' is a valid allocated vector
 *      'scalar' is a valid number
 * Post: 'vect' = 'vect' / 'scalar'
 */
void VectorDivideScalar(vector *vect, double scalar);

/* -- Function VectorMakeCopy
 * Pre: 'vect' is a valid allocated vector
 * Post: (a copy of 'vect' was allocated and returned ) ||
 *       ( there was an error with allocation && NULL was returned)
 */
vector *VectorMakeCopy(vector *vect);

/* -- Function VectorMakeHomogeneous
 * Pre: 'vect' is a valid allocated vector
 * Post: (a new vector was created containing ['vect', 1.0]) ||
 *       (there was an error with allocation && NULL was returned)
 */
vector *VectorMakeHomogeneous(vector *vect);

/* -- Function VectorMaxComponent
 * Pre: 'vect' is a valid allocated vector
 * Post: the 0-offset index of the maximum component of 'vect' was returned
 * Note: in the case of a tie for maximum, one of the indicies is returned
 */
int VectorMaxComponent(vector *vect);

/* -- Function VectorIdentical
 * Pre: 'vect1' and 'vect2' are valid allocated vectors
 * Post: (all components of 'vect1' are the same as 'vect2' && TRUE was returned) ||
 *       (FALSE was returned)
 */
bool VectorIdentical(vector *vect1, vector *vect2);

/* -- Function VectorManhattanEntropy
 * Pre: 'vect1' and 'vect2' are valid allocated vectors
 *      dim('vect1') == dim('vect2')
 * Post: The Manhattan Distance Entropy for 'vect1' and 'vect2' was
 *       calculated and returned
 * Note: H = Sum over dimensions(fd ln(fd))
 *       Where fd = |ad - bd| / Sum over d(|ad - bd|)
 */
double VectorManhattanEntropy(vector *vect1, vector *vect2);

/* -- Function VectorCorrelation
 * Pre: 'vect' is a valid allocated vector
 *      'matrixSize' points to an allocated integer
 * Post: The correlation matrix of 'vect' is returned as a 2D array of double.
 *       The matrix is square; the size of the matrix is placed in 'matrixSize'.
 *       NULL is returned if allocation is not possible.  In this case,
 *       'matrixSize' is undefined.
 *
 * Note: 'vect' = [a b c]
 *
 *           R = [a][a b c]
 *               [b]
 *               [c]
 *
 *             = [a2 ab ac]
 *               [ab b2 bc]
 *               [ac bc c2]
 */
double **VectorCorrelation(vector *vect, int *matrixSize);


#endif /* __vector_utils_h */


/* --- END of vector_utils.h --- */
